###################################################################################################
# _zcatalogmanager.py
#
# $Id$
# $Name$
# $Author$
# $Revision$
#
# Implementation of classes ZCatalogItem and ZCatalogManager (see below).
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###################################################################################################


# Imports.
from Products.ZCatalog import ZCatalog, Vocabulary, CatalogAwareness
# Product Imports.
import _globals


# -------------------------------------------------------------------------------------------------
#  _zcatalogmanager.zcat_meta_types
# -------------------------------------------------------------------------------------------------
zcat_meta_types = ['ZMS','ZMSRubrik','ZMSDocument','ZMSCustom','ZMSSysFolder','ZMSFile']

# -------------------------------------------------------------------------------------------------
#  _zcatalogmanager.search_string:
#
#  Search string of given value.
# -------------------------------------------------------------------------------------------------
def search_string(v):
  s = ''
  if v is not None:
    if type(v) is type('') and len(v) > 0:
      s += v + ' '
    elif type(v) is type([]):
      for i in v:
        s += search_string(i)
    elif type(v) is type({}):
      for k in v.keys():
        i = v[k]
        s += search_string(i)
  return s

# -------------------------------------------------------------------------------------------------
#  _zcatalogmanager.search_quote:
#
#  Remove HTML-Tags from given string.
# -------------------------------------------------------------------------------------------------
def search_quote(s, maxlen=255, tag='&middot;'):
  # remove all tags.
  stop = 0
  while not stop:
    iStartTag = s.find('<')
    iEndTag = s.find('>',iStartTag)
    if iStartTag >= 0 and iEndTag > iStartTag:
      s = s[:iStartTag] + tag + s[iEndTag+1:]
    else:
      stop = 1
  # limit characters.
  if len(s) > maxlen:
    blank = s.find(' ',maxlen)
    s = s[:blank] + '&middot;&middot;&middot;'
  # search quote.
  while s.find('>') >= 0:
    i = s.find('>')
    s = s[:i] + '&gt;' + s[i+1:]
  while s.find('<') >= 0:
    i = s.find('<')
    s = s[:i] + '&lt;' + s[i+1:]
  # return quoted search string.
  return s

# -------------------------------------------------------------------------------------------------
#  _zcatalogmanager._synchronizeCatalogText
# -------------------------------------------------------------------------------------------------
def _synchronizeCatalogText(self, REQUEST):
  if self.meta_type == 'ZMSTrashcan':
    return
  for lang in self.getLanguages():
    req = {'lang':lang,'manage_lang':self.getManageLanguage(lang)}
    v = ''
    if self.isPage():
      v = self.catalogText(req)
    setattr(self,'zcat_text_%s'%lang,v)
    v = ''
    if self.meta_type=='ZMSFile':
      v = self.catalogData(req)
    setattr(self,'zcat_data_%s'%lang,v)
  if self.isPage():
    for child in self.getChildNodes(REQUEST,zcat_meta_types):
      _synchronizeCatalogText(child,REQUEST)


###################################################################################################
###################################################################################################
###
###   class ZCatalogItem
###
###################################################################################################
###################################################################################################
class ZCatalogItem(CatalogAwareness.CatalogAware): 

    # ---------------------------------------------------------------------------------------------
    #	ZCatalogItem.search_quote:
    #
    #	Remove HTML-Tags.
    # ---------------------------------------------------------------------------------------------
    def search_quote(self, s, maxlen=255, tag="&middot;"):
      return search_quote(s,maxlen,tag)

    
    # ---------------------------------------------------------------------------------------------
    #	ZCatalogItem.getCatalogNavUrl:
    #
    #	Returns catalog-navigation url.
    # ---------------------------------------------------------------------------------------------
    def getCatalogNavUrl(self, REQUEST):
      return self.url_inherit_params(REQUEST['URL'],REQUEST,['qs'])


    # ---------------------------------------------------------------------------------------------
    #  ZCatalogItem.catalogText:
    #
    #  Catalog text.
    # ---------------------------------------------------------------------------------------------
    def catalogText(self, REQUEST):
      #++ if _globals.debug: print '[%s.catalogText]: %s'%(self.meta_type,self.absolute_url())
      s = ''
      if self.isVisible(REQUEST):
        for key in self.getObjAttrs().keys():
          obj_attr = self.getObjAttr(key)
          if obj_attr['xml']:
            datatype = obj_attr['datatype_key']
            if datatype in _globals.DT_STRINGS or \
               datatype == _globals.DT_DICTIONARY or \
               datatype == _globals.DT_LIST:
              value = self.getObjAttrValue(obj_attr,REQUEST)
              value = search_string(value)
              # Increase weight of title-attributes!
              if obj_attr['key'] in ['title','titlealt','titleshort']:
                value = value + value + value
              s += value
        for ob in self.getChildNodes(REQUEST,self.PAGEELEMENTS):
          s += ob.catalogText(REQUEST)
      return s


    # ---------------------------------------------------------------------------------------------
    #  ZCatalogItem.catalogData:
    #
    #  Catalog data.
    # ---------------------------------------------------------------------------------------------
    def catalogData(self, REQUEST):
      v = ''
      if self.isVisible(REQUEST):
        file = self.getObjProperty('file',REQUEST)
        if file is not None:
          v = file.getData()
      return v


    # ---------------------------------------------------------------------------------------------
    #  ZCatalogItem.synchronizeSearch:
    # ---------------------------------------------------------------------------------------------
    def synchronizeSearch(self, REQUEST):
      if self.getConfProperty('ZMS.CatalogAwareness.active',1)==0: 
        return

      #-- Reindex
      try:
        lang = REQUEST.get('lang',self.getPrimaryLanguage())
        #-- Page.
        v = ''
        ob = self
        while ob is not None and not ob.isPage():
          ob = ob.getParentNode()
        v = ob.catalogText(REQUEST)
        setattr(ob,'zcat_text_%s'%lang,v)
        #-- File.
        v = ''
        ob = self
        if ob.meta_type=='ZMSFile':
          v = ob.catalogData(REQUEST)
        setattr(ob,'zcat_data_%s'%lang,v)
        #-- Reindex.
        #++ if _globals.debug: print "[%s.synchronizeSearch]: Reindex %s."%(self.meta_type,self.id)
        ob = self
        while ob is not None and not (ob.isPage() or ob.meta_type=='ZMSFile'):
          ob = ob.getParentNode()
        ob.default_catalog = 'catalog'
        ob.reindex_object()
      except:
        pass


###################################################################################################
###################################################################################################
###
###   class ZCatalogManager
###
###################################################################################################
###################################################################################################
class ZCatalogManager:

    # ---------------------------------------------------------------------------------------------
    #  ZCatalogManager.getCatalog:
    #
    #  Returns catalog.
    # ---------------------------------------------------------------------------------------------
    def getCatalog(self):
      context = self.getDocumentElement()
      obs = context.objectValues(['ZCatalog'])
      if len(obs) == 0:
        return self.recreateCatalog()
      return obs[0]


    # ---------------------------------------------------------------------------------------------
    #  ZCatalogManager.recreateCatalog:
    #
    #  Recreates catalog.
    # ---------------------------------------------------------------------------------------------
    def recreateCatalog(self):
      message = ''
      context = self.getDocumentElement()
      
      #-- Delete catalog
      ids = context.objectIds(['ZCatalog'])
      if len(ids) > 0:
        context.manage_delObjects(ids)
      
      #-- (Re-)create catalog
      cat_id = 'catalog'
      cat_title = 'Catalog'
      vocab_id = 'create_default_catalog_'
      zcatalog = ZCatalog.ZCatalog(cat_id, cat_title, vocab_id, context)
      context._setObject(zcatalog.id, zcatalog)
      zcatalog = self.getCatalog()
      
      #-- Remove Default-Vocabulary
      ids = zcatalog.objectIds(['Vocabulary'])
      if len(ids) > 0:
        zcatalog.manage_delObjects(ids,None)
      
      #-- Add ISO 8859-1 Vocabulary
      # see ZCatalogs with Umlauts
      # http://www.zope.org/Members/strobl/HowTos/Iszcatalog
      id = "Vocabulary"
      title = id
      globbing = 1
      splitter = "ISO_8859_1_Splitter"
      oVocabulary = Vocabulary.Vocabulary(id,title,globbing,splitter)
      zcatalog._setObject(oVocabulary.id,oVocabulary)
      
      #-- Clear catalog
      zcatalog.manage_catalogClear()
      
      #-- Delete indexes from catalog
      zcatalog.manage_delColumn(zcatalog.schema())
      zcatalog.manage_delIndex(zcatalog.indexes())
      
      #-- (Re-)create indexes on catalog
      index_types = []
      for index in zcatalog.Indexes.filtered_meta_types():
        index_types.append(index['name'])
      for lang in self.getLanguages():
        message += "Index ["
        index_name = 'zcat_text_%s'%lang
        index_type = 'TextIndex'
        index_extras = None
        zcatalog.manage_addColumn(index_name)
        zcatalog.manage_addIndex(index_name,index_type,index_extras)
        message += index_name
        index_type = None
        for k in ['TextIndexNG2']:
          if k in index_types: index_type = k
        if self.getConfProperty('ZCatalog.TextIndexNG',0)==1 and index_type is not None:
          index_name = 'zcat_data_%s'%lang
          index_extras = _globals.MyClass()
          setattr(index_extras,'default_encoding','iso-8859-15')
          setattr(index_extras,'indexed_fields',index_name)
          setattr(index_extras,'near_distance',5)
          setattr(index_extras,'splitter_casefolding',1)
          setattr(index_extras,'splitter_max_len',64)
          setattr(index_extras,'splitter_separators','.+-_@')
          setattr(index_extras,'splitter_single_chars',0)
          setattr(index_extras,'use_converters',1)
          setattr(index_extras,'use_normalizer','')
          setattr(index_extras,'use_stemmer','')
          setattr(index_extras,'use_stopwords','')
          zcatalog.manage_addIndex(index_name,index_type,index_extras)
          message += ", "+index_name
        message += "] created for language <i>"+self.getLanguageLabel(lang)+'</i><br>'
      
      #-- Return message.
      return message


    # ---------------------------------------------------------------------------------------------
    #  ZCatalogManager.reindexCatalog:
    #
    #  Reindex catalog.
    # ---------------------------------------------------------------------------------------------
    def reindexCatalog(self, REQUEST):
      message = ''
      
      #-- Recreate catalog.
      message += self.recreateCatalog()
      zcatalog = self.getCatalog()
      
      #-- Push all ZMS contents into single full-text attribute zcat_text_$lang
      _synchronizeCatalogText(self,REQUEST)
      
      #-- Find items to catalog.
      homeElmnt = self.getHome()
      path = homeElmnt.absolute_url()[len(str(REQUEST['BASE0'])):]
      results = zcatalog.ZopeFindAndApply(homeElmnt,
                 obj_metatypes=zcat_meta_types,
                 obj_ids=None,
                 obj_searchterm=None,
                 obj_expr=None,
                 obj_mtime=None,
                 obj_mspec=None,
                 obj_permission=None,
                 obj_roles=None,
                 search_sub=1,
                 REQUEST=REQUEST,
                 apply_func=zcatalog.catalog_object,
                 apply_path=path)
      
      # Return with message.
      message += 'Catalog indexed successfully.'
      return message


    # ---------------------------------------------------------------------------------------------
    #	ZCatalogManager.getCatalogQueryString:
    # ---------------------------------------------------------------------------------------------
    def getCatalogQueryString(self, raw, option='AND'):
      qs = ''
      raw = raw.replace(' and ',' AND ')
      raw = raw.replace(' or ',' OR ')
      if raw.find(' AND ')>=0 and raw.find(' OR ')>=0:
        for raw_item in raw.split(' '):
          if len(raw_item)>0:
            if raw_item in ['AND','OR'] or raw_item[-1]=='*':
              qs += raw_item + ' '
            else:
              qs += raw_item + '* '
      else:
        q = 0
        c = 0
        for raw_item in raw.split(' '):
          if len(raw_item)>0:
            if q==0 and c>0:
              qs += option + ' '
            if raw_item[0]=='*':
              q=1
            if q==1:
              qs += raw_item +' '
            else:
              qs += raw_item + '* '
            if raw_item[-1]=='*':
              q=0
            c=c+1
      qs = qs.replace('-','* AND *')
      qs = qs.strip()
      return qs


    # ---------------------------------------------------------------------------------------------
    #	ZCatalogManager.submitCatalogQuery:
    #
    #	Submits query to catalog.
    # ---------------------------------------------------------------------------------------------
    def submitCatalogQuery(self, search_query, search_order_by, search_meta_types, REQUEST):
    
      #-- Numbers are not cataloged.
      lnum = []
      qs = ''
      lq = search_query.split(' ')
      for i in range(len(lq)):
        try: 
          lnum.append(int(lq[i].strip()))
        except:
          qs += lq[i] + ' '
      
      #-- Search catalog.
      lang = REQUEST['lang']
      zcatalog = self.getCatalog()
      items = []
      for zcindex in zcatalog.indexes():
        if zcindex.find('zcat_')==0 and zcindex.rfind('_'+lang)==len(zcindex)-len('_'+lang):
          d = {}
          d['meta_type'] = zcat_meta_types
          d[zcindex] = qs
          items.extend(zcatalog(d))
      
      #-- Sort order.
      if int(search_order_by)==1:
        order_by = 'score'
      else:
        order_by = 'time'
      
      #-- Process results.
      results = []
      for item in items:
        # Get object by path.
        data_record_id = item.data_record_id_
        url = zcatalog.getpath(data_record_id)
        ob = self.getHome()
        i = url.find(ob.id)
        if i >= 0:
          i += len(ob.id)
        else:
          i += 1
        for id in url[i+1:].split('/'):
          if ob is not None:
            ob = getattr(ob,id,None)
        # Uncatalog object.
        uncatalog = ob is None
        if not uncatalog:
          for path_ob in ob.breadcrumbs_obj_path():
            if not uncatalog:
              uncatalog = uncatalog or path_ob == self.getTrashcan()
              uncatalog = uncatalog or not path_ob.isVisible(REQUEST)
        if uncatalog:
          try:
            zcatalog.uncatalog_object(url)
          except:
            pass
          ob = None
        # Check for valid result.
        append = ob is not None
        if append:
          # Handle Pages.
          append = ob.isPage() or ob.meta_type=='ZMSFile'
          # Handle Meta-Types.
          if len(search_meta_types) > 0:
            if ob.meta_type == 'ZMSCustom':
              append = append and ob.meta_id in search_meta_types
            else:
              append = append and ob.meta_type in search_meta_types
          # Handle Numbers.
          zctext = getattr(ob,'zcat_text_%s'%lang,'')
          for num in lnum:
            append = append and zctext.find(str(num))>=0
        # Append to valid results.
        if append:
          sum_time = ob.getObjProperty('attr_dc_date',REQUEST)
          if sum_time is None or len(str(sum_time))==0:
            sum_time = ob.getObjProperty('change_dt',REQUEST)
          try:
            sum_score = int(item.data_record_score_)
          except:
            sum_score = 0
          if ob.meta_type == 'ZMSFile':
            id = ob.id+'/'
            sum_url = ob.getObjProperty('file',REQUEST).getHref(REQUEST)
            sum_url = sum_url[sum_url.find(id)+len(id):]
          else:
            sum_url = 'index_%s.html'%lang
            if REQUEST.get('preview','')=='preview':
              sum_url = self.url_append_params(sum_url,{'preview':'preview'})
            if ob.meta_type=='ZMSCustom':
              sum_url = self.url_append_params(sum_url,{'ZMS_PARAM0':'search','ZMS_PARAM1':search_query})
          result = {}
          result['ob'] = ob
          result['title'] = ob.getTitle(REQUEST)
          result['description'] = ob.getDCDescription(REQUEST)
          result['score'] = sum_score
          result['time'] = sum_time
          result['url'] = ob.absolute_url() + '/' + sum_url
          results.append((result[order_by],result))
      
      #-- Sort objects.
      results.sort()
      results.reverse()
      
      #-- Return objects in correct sort-order.
      return map(lambda ob: ob[1],results)

###################################################################################################
